"""
Method 4: Base64 Groups

Encodes data as base64 and stores in the groups property.

Objects inside levels can have non-numeric groups but Geometry Dash doesn't load them when playing
But that's not really relevant - we just want the level data to be preserved

T̶h̶i̶s̶ ̶c̶o̶u̶l̶d̶ ̶w̶o̶r̶k̶ ̶b̶u̶t̶ ̶I̶'̶m̶ ̶u̶n̶s̶u̶r̶e̶ ̶t̶h̶e̶ ̶l̶e̶v̶e̶l̶ ̶i̶s̶ ̶a̶c̶t̶u̶a̶l̶l̶y̶ ̶p̶r̶e̶s̶e̶r̶v̶e̶d̶
It is not preserved - it strips any non-numeric groups
"""
from gdparse import GDLevel, LevelObject
from pathlib import Path
import base64
from .compression import compress_data, decompress_data

BLOCK_ID = 211
CHUNK_SIZE = 9999


def encode(filepath: str | Path, skip_compression: bool = False) -> str:
    # Objects inside levels can have non-numeric groups but Geometry Dash doesn't load them when playing
    # But that's not really relevant - we just want the level data to be preserved
    filepath = Path(filepath)
    data = filepath.read_bytes()
    if not skip_compression:
        data=compress_data(data)
    b64_data=base64.b64encode(data).decode('ascii')
    chunks = [b64_data[i:i+CHUNK_SIZE] for i in range (0, len(b64_data), CHUNK_SIZE)]
    level = GDLevel.create_empty()
    for i, chunk in enumerate(chunks):
        obj = LevelObject.create_block(block_id=BLOCK_ID, x=i*10, y=0)
        obj.properties[57]= chunk
        level.add_object(obj)
    return level.serialize()

# This could work but I'm unsure the level is actually preserved
# It is not preserved - it strips any non-numeric groups


def decode(level_string: str, skip_decompression: bool = False) -> bytes:
    level = GDLevel(level_string)

    # Concatenate all chunks
    b64_string = ''.join(obj.properties.get(57) for obj in level.objects)
    data = base64.b64decode(b64_string)
    if not skip_decompression:
        data = decompress_data(data)

    return data

# ^ This is for the group method using Base64



"""
Method 2: Raw Groups

Stores raw byte values (0-255) directly as group numbers.
Each object can have up to 9999 groups, so we chunk the data.

This worked! YAY - But the level was ~1GB, and while it did load
Geometry Dash took longer to load in general and the game crashed when you tried to close it
But the level preserved! So technically it worked...?
Obviously storing raw data wasn't gonna be space or time efficient
"""
from gdparse import GDLevel, LevelObject
from pathlib import Path
from .compression import compress_data, decompress_data

BLOCK_ID = 211
CHUNK_SIZE = 9999


def encode(filepath: str | Path, skip_compression: bool = False) -> str:
    # This one might work... Geometry Dash doesn't check
    # for repeating groups inside objects when loading the level
    filepath = Path(filepath)
    data = filepath.read_bytes()
    if not skip_compression:
        data=compress_data(data)
    chunks = [data[i:i+CHUNK_SIZE] for i in range (0, len(data), CHUNK_SIZE)]
    level = GDLevel.create_empty()
    for i, chunk in enumerate(chunks):
        #Convert chunk bytes to groups string
        groups_str = '.'.join(str(b) for b in chunk)
        obj = LevelObject.create_block(block_id=BLOCK_ID, x=i*10, y=0)
        obj.properties[57]=groups_str
        level.add_object(obj)

    return level.serialize()

# ^ This worked! YAY - But the level was ~1GB, and while it did load
# Geometry Dash took longer to load in general and the game crashed when you tried to close it
# But the level preserved! So technically it worked...?
# Obviously storing raw data wasn't gonna be space or time efficient


def decode(level_string: str, skip_decompression: bool = False) -> bytes:
    level = GDLevel(level_string)

    group_data=[]

    for obj in level.objects:
        object_groups=obj.properties.get(57).split('.')
        byte_values= [int(x) for x in object_groups]
        group_data.extend(byte_values)
    data = bytes(group_data)
    if not skip_decompression:
        data = decompress_data(data)

    return data

# ^ This is for the group method



"""
Method 3: Base 10000

Uses base 10000 encoding to make better use of the 9999 available groups.
Processes 256 bytes at a time and converts to base 10000.

So this helps - not by a lot but it does make the level smaller - but not small enough
Also - the time to process this is insanely long - simply not worth it

Python big int division is O(n^2), so smaller chunks = faster
"""
from gdparse import GDLevel, LevelObject
from pathlib import Path
from .compression import compress_data, decompress_data

BLOCK_ID = 211
BASE10000_CHUNK = 256  # bytes per chunk


def encode(filepath: str | Path, skip_compression: bool = False) -> str:
    # Let's try using Base 10000
    # There are 9999 groups - we only used the first 255
    filepath = Path(filepath)
    data = filepath.read_bytes()
    if not skip_compression:
        data=compress_data(data)
    level = GDLevel.create_empty()
    for i in range(0, len(data), BASE10000_CHUNK):
        chunk = data[i:i + BASE10000_CHUNK]

        # Convert Bytes to Base 10000
        big_int = int.from_bytes(chunk, 'big')
        groups=[]
        while big_int > 0:
            groups.append(big_int % 10000)
            big_int //= 10000
        groups = list(reversed(groups)) or [0]

        #Store chunk length as first group
        groups = [len(chunk)] + groups

        groups_str = '.'.join(str(g) for g in groups)
        obj = LevelObject.create_block(block_id=BLOCK_ID, x=i*30, y=0)
        obj.properties[57] = groups_str
        level.add_object(obj)

    return level.serialize()


# So this helps - not by a lot but it does make the level smaller - but not small enough
# Also - the time to process this is insanely long - simply not worth it


def decode(level_string: str, skip_decompression: bool = False) -> bytes:
    level = GDLevel(level_string)
    result = b''

    for obj in level.objects:
        groups = [int(g) for g in obj.properties.get(57).split('.')]
        chunk_len = groups[0]
        groups = groups[1:]

        big_int = 0
        for g in groups:
            big_int = big_int * 10000 + g

        result += big_int.to_bytes(chunk_len, 'big')

    if not skip_decompression:
        result = decompress_data(result)

    return result

# ^ This is for the group method using Base10000 for storing



from .compression import compress_data, decompress_data
from .method1_xy import encode as method1_encode, decode as method1_decode
from .method2_raw_groups import encode as method2_encode, decode as method2_decode
from .method3_base10000 import encode as method3_encode, decode as method3_decode
from .method4_base64_groups import encode as method4_encode, decode as method4_decode
from .method5_property31 import encode as method5_encode, decode as method5_decode
from .method6_optimized import encode as method6_encode, decode as method6_decode

METHODS = {
    1: (method1_encode, method1_decode, "X/Y Coordinates - Unoptimized"),
    2: (method2_encode, method2_decode, "Raw Groups - 1GB levels"),
    3: (method3_encode, method3_decode, "Base 10000 - Slow"),
    4: (method4_encode, method4_decode, "Base64 Groups - Stripped by GD"),
    5: (method5_encode, method5_decode, "Property 31 - Doesn't work"),
    6: (method6_encode, method6_decode, "Optimized Base 9999 - Best"),
}

DEFAULT_METHOD = 6



import zstandard as zstd

def compress_data(data: bytes) -> bytes:
    compressor = zstd.ZstdCompressor(level=19)
    return compressor.compress(data)


def decompress_data(data: bytes) -> bytes:
    decompressor = zstd.ZstdDecompressor()
    return decompressor.decompress(data)



"""
GD Storage - Encode files into Geometry Dash levels
"""
import argparse
import sys
import os
import base64
import gzip
import getpass
import re
import json
import platform
from pathlib import Path

from methods import METHODS


# Config file location
CONFIG_DIR = Path.home() / ".config" / "gd-storage"
CONFIG_FILE = CONFIG_DIR / "config.json"


def load_config() -> dict:
    """Load config from file or return defaults."""
    if CONFIG_FILE.exists():
        try:
            return json.loads(CONFIG_FILE.read_text())
        except (json.JSONDecodeError, OSError):
            pass
    return {}


def save_config(config: dict):
    """Save config to file."""
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    CONFIG_FILE.write_text(json.dumps(config, indent=2))


def get_default_gd_path() -> str | None:
    """Get default GD save path for current platform."""
    system = platform.system()
    if system == "Windows":
        local_app_data = os.getenv("LOCALAPPDATA")
        if local_app_data:
            return os.path.join(local_app_data, "GeometryDash")
    elif system == "Darwin":  # macOS
        return os.path.expanduser("~/Library/Application Support/GeometryDash")
    return None


def get_manager(config: dict):
    """Get GD save manager based on config."""
    from save_manager import new_manager

    gd_path = config.get("gd_path")
    gd_format = config.get("format", "auto")

    # Auto-detect path if not configured
    if not gd_path:
        gd_path = get_default_gd_path()
        if not gd_path:
            raise ValueError("GD path not configured. Run: gd-storage --config")

    # Check path exists
    if not os.path.exists(gd_path):
        raise ValueError(f"GD save folder not found: {gd_path}\nRun: gd-storage --config")

    return new_manager(path=gd_path, format=gd_format)


def cmd_config():
    """Configure GD storage settings."""
    config = load_config()

    print("GD Storage Configuration")
    print("=" * 40)

    # Show current settings
    current_path = config.get("gd_path", get_default_gd_path() or "Not set")
    current_format = config.get("format", "auto")
    print(f"Current path: {current_path}")
    print(f"Current format: {current_format}")
    print()

    # Get new path
    print("Enter GD save folder path (or press Enter to keep current):")
    print("  Windows: %LOCALAPPDATA%\\GeometryDash")
    print("  macOS: ~/Library/Application Support/GeometryDash")
    print("  Linux/Proton: ~/.steam/steam/steamapps/compatdata/.../GeometryDash")
    new_path = input("> ").strip()

    if new_path:
        new_path = os.path.expanduser(os.path.expandvars(new_path))
        if not os.path.exists(new_path):
            print(f"Warning: Path does not exist: {new_path}")
            confirm = input("Save anyway? (y/N): ").strip().lower()
            if confirm != 'y':
                return 1
        config["gd_path"] = new_path

    # Get format
    print()
    print("Save format:")
    print("  1. auto (detect from platform)")
    print("  2. windows (XOR encryption - Windows/Linux/Proton)")
    print("  3. mac (AES encryption - macOS)")
    choice = input("Choice [1]: ").strip() or "1"

    if choice == "2":
        config["format"] = "windows"
    elif choice == "3":
        config["format"] = "mac"
    else:
        config["format"] = "auto"

    save_config(config)
    print()
    print(f"Config saved to {CONFIG_FILE}")
    return 0


def save_decoded_file(filename: str, data: bytes) -> Path | None:
    """Save decoded file to Downloads, checking for overwrites."""
    # Sanitize filename - prevent path traversal
    safe_filename = Path(filename).name
    if not safe_filename:
        safe_filename = "decoded_file"

    downloads = Path(os.path.expanduser("~")) / "Downloads" / safe_filename

    if downloads.exists():
        print(f"File already exists: {downloads}")
        choice = input("Overwrite? (y/N): ").strip().lower()
        if choice != 'y':
            # Try adding a number suffix
            stem = downloads.stem
            suffix = downloads.suffix
            for i in range(1, 100):
                alt = downloads.parent / f"{stem}_{i}{suffix}"
                if not alt.exists():
                    downloads = alt
                    break
            else:
                print("Could not find available filename")
                return None

    downloads.write_bytes(data)
    return downloads


def show_help():
    print("GD Storage - Encode files into Geometry Dash levels")
    print()
    print("Usage:")
    print("  gd-storage --upload <filepath>    Encode and upload to GD servers")
    print("  gd-storage --fetch <level_id>     Download and decode from GD servers")
    print("  gd-storage --encode <filepath>    Encode and inject into local GD save")
    print("  gd-storage --decode <levelname>   Decode from local GD save")
    print("  gd-storage --config               Configure GD save path")


def cmd_upload(filepath: Path, encode_func):
    """Encode and upload a file to GD servers."""
    import dashlib

    if not filepath.exists():
        print(f"File not found: {filepath}")
        return 1

    print(f"Encoding {filepath.name} ({filepath.stat().st_size:,} bytes)...")
    level_str = encode_func(filepath)

    # Decompress to get raw level for object count
    if level_str.startswith('H4sI'):
        compressed = base64.urlsafe_b64decode(level_str + '==')
        raw_level = gzip.decompress(compressed).decode('utf-8')
    else:
        raw_level = level_str

    obj_count = raw_level.count(';')
    level_name = filepath.stem[:20]
    description = f"Encoded: {filepath.name}"
    desc_encoded = base64.urlsafe_b64encode(description.encode()).decode()

    class UploadLevel:
        def __init__(self):
            self.levelString = level_str
            self.levelName = level_name
            self.description = desc_encoded
            self.password = 1
            self.copiedId = 0
            self.version = 1
            self.officialSong = 0
            self.customSongID = 0
            self.coins = 0
            self.lowDetailMode = False
            self.objects = obj_count
            self.isAuto = False
            self.twoPlayer = False
            self.starsRequested = 0
            self.editorTime = 0
            self.editorTimeCopies = 0
            self.length = dashlib.LENGTH_TINY

    level = UploadLevel()

    # Get credentials
    username, account_id, gjp2 = get_credentials()
    if not username:
        return 1

    print(f"Uploading '{level_name}'...")
    try:
        result = dashlib.uploadLevel(
            level=level,
            username=username,
            accountID=account_id,
            gjp2=gjp2,
        )
        if result == "-1" or result.startswith("-"):
            raise ValueError(f"Server returned: {result}")
        new_level_id = int(result)
        print(f"Uploaded! Level ID: {new_level_id}")
        print(f"Fetch with: gd-storage --fetch {new_level_id}")
        return 0
    except Exception as e:
        print(f"Upload failed: {e}")
        return 1


def cmd_fetch(level_id: int, decode_func):
    """Download and decode a level from GD servers."""
    from gd_api import download_level

    print(f"Fetching level {level_id}...")
    try:
        level_data = download_level(level_id)
    except Exception as e:
        print(f"Failed to fetch: {e}")
        return 1

    level_name = level_data.get("name", "Unknown")
    description = level_data.get("description", "")
    level_str = level_data.get("level_string", "")

    print(f"Level: {level_name}")
    if description:
        print(f"Description: {description}")

    try:
        filename, data = decode_func(level_str)
        saved_path = save_decoded_file(filename, data)
        if saved_path:
            print(f"Saved to {saved_path} ({len(data):,} bytes)")
            return 0
        return 1
    except Exception as e:
        print(f"Failed to decode: {e}")
        return 1


def cmd_encode(filepath: Path, encode_func):
    """Encode and inject into local GD save."""
    if not filepath.exists():
        print(f"File not found: {filepath}")
        return 1

    print(f"Encoding {filepath.name} ({filepath.stat().st_size:,} bytes)...")
    level_str = encode_func(filepath)

    config = load_config()
    try:
        manager = get_manager(config)
    except ValueError as e:
        print(f"Error: {e}")
        return 1

    manager.injectLevel(level_str, filepath.stem, f"Encoded: {filepath.name}")
    manager.save(ccll=True, ccgm=False)
    print(f"Injected as '{filepath.stem}'")
    return 0


def cmd_decode(level_name: str, decode_func):
    """Decode from local GD save."""
    print(f"Extracting '{level_name}'...")

    config = load_config()
    try:
        manager = get_manager(config)
    except ValueError as e:
        print(f"Error: {e}")
        return 1

    ccll = manager.ccll.decode('utf-8', errors='ignore')

    start = ccll.find(f'<s>{level_name}</s>')
    if start == -1:
        print(f"Level '{level_name}' not found!")
        return 1

    k4_start = ccll.find('<k>k4</k><s>', start) + len('<k>k4</k><s>')
    k4_end = ccll.find('</s>', k4_start)
    level_str = ccll[k4_start:k4_end]

    try:
        filename, data = decode_func(level_str)
        saved_path = save_decoded_file(filename, data)
        if saved_path:
            print(f"Saved to {saved_path} ({len(data):,} bytes)")
            return 0
        return 1
    except Exception as e:
        print(f"Failed to decode: {e}")
        return 1


def get_credentials():
    """Get GD credentials from save file or prompt user."""
    from gd_api import get_account_id

    config = load_config()
    try:
        manager = get_manager(config)
        ccgm = manager.ccgm.decode('utf-8', errors='ignore')

        saved_user = re.search(r'<k>GJA_001</k><s>([^<]+)</s>', ccgm)
        saved_id = re.search(r'<k>GJA_003</k><i>(\d+)</i>', ccgm)
        saved_gjp2 = re.search(r'<k>GJA_005</k><s>([^<]+)</s>', ccgm)

        if saved_user and saved_id and saved_gjp2:
            print(f"Using saved credentials: {saved_user.group(1)}")
            return saved_user.group(1), int(saved_id.group(1)), saved_gjp2.group(1)
    except ValueError:
        pass  # No local save, prompt for credentials

    print("Enter GD credentials:")
    username = input("Username: ")
    password = getpass.getpass("Password: ")

    try:
        account_id, gjp2 = get_account_id(username, password)
        print(f"Logged in as {username}")
        return username, account_id, gjp2
    except Exception as e:
        print(f"Login failed: {e}")
        return None, None, None


def main():
    parser = argparse.ArgumentParser(
        description="GD Storage - Encode files into Geometry Dash levels",
        add_help=False
    )
    parser.add_argument('--upload', metavar='FILE', help='Encode and upload to GD servers')
    parser.add_argument('--fetch', metavar='ID', type=int, help='Download and decode from GD servers')
    parser.add_argument('--encode', metavar='FILE', help='Encode and inject into local GD save')
    parser.add_argument('--decode', metavar='NAME', help='Decode from local GD save')
    parser.add_argument('--config', action='store_true', help='Configure GD save path')
    parser.add_argument('--help', '-h', action='store_true', help='Show help')

    args = parser.parse_args()

    # Handle config
    if args.config:
        return cmd_config()

    # Show help if no args or --help
    if args.help or (not args.upload and not args.fetch and not args.encode and not args.decode):
        show_help()
        return 0

    encode_func, decode_func, _ = METHODS[6]

    # Run command
    if args.upload:
        return cmd_upload(Path(args.upload), encode_func)
    elif args.fetch:
        return cmd_fetch(args.fetch, decode_func)
    elif args.encode:
        return cmd_encode(Path(args.encode), encode_func)
    elif args.decode:
        return cmd_decode(args.decode, decode_func)

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Geometry Dash Server API

Raw HTTP implementation for downloading and uploading levels.
"""
import urllib.request
import urllib.parse
import base64
import hashlib
import zlib

GD_URL = "https://www.boomlings.com/database"
SECRET = "Wmfd2893gb7"  # Public secret used by GD


def download_level(level_id: int) -> dict:
    """
    Download a level from GD servers by ID.
    Returns dict with 'level_string', 'name', 'description', etc.
    """
    data = urllib.parse.urlencode({
        "levelID": level_id,
        "secret": SECRET,
        "gameVersion": 22,
        "binaryVersion": 42,
        "gdw": 0,
        "inc": 1,
        "extras": 0,
    }).encode()

    req = urllib.request.Request(f"{GD_URL}/downloadGJLevel22.php", data=data)
    req.add_header("User-Agent", "")

    with urllib.request.urlopen(req, timeout=30) as response:
        result = response.read().decode('utf-8', errors='ignore')


    if result == "-1":
        raise ValueError(f"Level {level_id} not found")

    # Parse the response (key:value:key:value format)
    parts = result.split("#")[0]  # Remove hash/creator info
    fields = parts.split(":")
    level_data = {}
    for i in range(0, len(fields) - 1, 2):
        level_data[fields[i]] = fields[i + 1]

    # Decode the level string (base64 + gzip)
    level_string_encoded = level_data.get("4", "")
    if level_string_encoded:
        try:
            # URL-safe base64 decode
            decoded = base64.urlsafe_b64decode(level_string_encoded + "==")
            # Gzip decompress
            level_string = zlib.decompress(decoded, 15 + 32).decode('utf-8', errors='ignore')
        except (ValueError, zlib.error):
            # Some levels might not be compressed
            level_string = level_string_encoded
        level_data["level_string"] = level_string

    # Decode description (base64)
    desc_encoded = level_data.get("3", "")
    if desc_encoded:
        try:
            level_data["description"] = base64.urlsafe_b64decode(desc_encoded + "==").decode('utf-8', errors='ignore')
        except (ValueError, UnicodeDecodeError):
            level_data["description"] = desc_encoded

    # Key mappings for convenience
    level_data["name"] = level_data.get("2", "Unknown")
    level_data["id"] = level_data.get("1", level_id)

    return level_data


def gjp_encode(password: str) -> str:
    """
    Encode password for GD API (old GJP format).
    XOR with key then base64.
    """
    key = "37526"
    xored = ''.join(chr(ord(c) ^ ord(key[i % len(key)])) for i, c in enumerate(password))
    return base64.urlsafe_b64encode(xored.encode('latin-1')).decode()


def gjp2_encode(password: str) -> str:
    """
    Encode password for GD API (GJP2 format).
    SHA1 hash + salt.
    """
    salted = password + "mI29fmAnxgTs"
    hashed = hashlib.sha1(salted.encode()).hexdigest()
    return hashed


def upload_level(
    username: str,
    gjp2: str,
    account_id: int,
    level_name: str,
    level_string: str,
    description: str = "",
    unlisted: bool = True,
) -> int:
    """
    Upload a level to GD servers.
    Returns the new level ID.

    Note: This function is kept for reference but dashlib.uploadLevel is preferred.
    """

    # Decompress to get raw level string for obj count
    if level_string.startswith('H4sI'):
        compressed_bytes = base64.urlsafe_b64decode(level_string + '==')
        raw_level = zlib.decompress(compressed_bytes, 15 + 32).decode('utf-8')
    else:
        raw_level = level_string

    level_encoded = level_string
    obj_count = raw_level.count(';')

    # Encode description
    desc_encoded = base64.urlsafe_b64encode(description.encode()).decode()

    # Level length based on object count (0=Tiny, 1=Short, 2=Medium, 3=Long, 4=XL)
    if obj_count > 10000:
        level_length = 4  # XL
    elif obj_count > 5000:
        level_length = 3  # Long
    elif obj_count > 1000:
        level_length = 2  # Medium
    elif obj_count > 200:
        level_length = 1  # Short
    else:
        level_length = 0  # Tiny

    # Level upload data
    data = urllib.parse.urlencode({
        "accountID": account_id,
        "gjp2": gjp2,
        "userName": username,
        "levelName": level_name,
        "levelDesc": desc_encoded,
        "levelVersion": 1,
        "levelLength": level_length,
        "audioTrack": 0,
        "auto": 0,
        "password": 1,  # No copy
        "original": 0,
        "twoPlayer": 0,
        "songID": 0,
        "objects": obj_count,
        "coins": 0,
        "requestedStars": 0,
        "unlisted": 1 if unlisted else 0,
        "wt": 0,
        "wt2": 0,
        "ldm": 0,
        "levelString": level_encoded,
        "seed": "abc123",
        "seed2": generate_seed2(level_encoded),
        "secret": SECRET,
        "gameVersion": 22,
        "binaryVersion": 44,
        "gdw": 0,
    }).encode()

    req = urllib.request.Request(f"{GD_URL}/uploadGJLevel21.php", data=data)
    req.add_header("User-Agent", "")

    with urllib.request.urlopen(req, timeout=60) as response:
        result = response.read().decode()

    if result == "-1":
        raise ValueError("Upload failed - invalid credentials or verification")
    elif result.startswith("-"):
        raise ValueError(f"Upload failed with error code: {result}")

    return int(result)


def xor_cipher(inp: str, key: str) -> str:
    """XOR cipher with cyclic key."""
    result = ""
    for i, char in enumerate(inp):
        result += chr(ord(char) ^ ord(key[i % len(key)]))
    return result


def generate_upload_seed(data: str, chars: int = 50) -> str:
    """Generate upload seed from level string - take every Nth char to get 50 chars."""
    if len(data) < chars:
        return data
    step = len(data) // chars
    return data[::step][:chars]


def generate_seed2(level_string: str) -> str:
    """Generate seed2 for level upload verification (CHK format)."""
    # Get 50 evenly-spaced characters from level string
    seed_chars = generate_upload_seed(level_string)
    # Add salt and hash
    salt = "xI25fpAapCQg"
    to_hash = seed_chars + salt
    hashed = hashlib.sha1(to_hash.encode()).hexdigest()
    # XOR with key and base64 encode
    xored = xor_cipher(hashed, "41274")
    return base64.urlsafe_b64encode(xored.encode()).decode()


def get_account_id(username: str, password: str) -> tuple[int, str]:
    """
    Login and get account ID.
    Returns (account_id, gjp2) for use in subsequent requests.
    """
    # Login uses a different secret than other endpoints
    login_secret = "Wmfv3899gc9"

    data = urllib.parse.urlencode({
        "userName": username,
        "password": password,  # Raw password for login
        "secret": login_secret,
        "udid": "S15232137420643451451521515121125115195140311",
        "gameVersion": 22,
        "binaryVersion": 42,
    }).encode()

    req = urllib.request.Request(f"{GD_URL}/accounts/loginGJAccount.php", data=data)
    req.add_header("User-Agent", "")
    req.add_header("Content-Type", "application/x-www-form-urlencoded")

    with urllib.request.urlopen(req, timeout=30) as response:
        result = response.read().decode()

    if result == "-1" or result.startswith("-"):
        raise ValueError(f"Login failed (error: {result})")

    # Returns accountID,userID
    parts = result.split(",")
    account_id = int(parts[0])

    # Generate GJP2 for subsequent requests
    gjp2 = gjp2_encode(password)

    return account_id, gjp2



"""
Method 6: Optimized Base 9999 (THE BEST ONE)

Uses base 9999 encoding with 8-byte chunks for fast processing.
- Processes 8 bytes at a time (avoids O(n^2) big int division)
- Uses groups 1-9999
- Stores original data length to handle padding
- Each 8 bytes becomes exactly 5 groups

8 byte processing is way faster
Faster rendering & smaller (~13%)
Also, I forgot group 0 doesn't exist. I don't really use the GD Editor...

Results:
- Level string overhead: ~2.95x
- On-disk overhead: ~1.73x (GD compresses the save file)

What else was fixed in this version:
- Objects can have at most 10 groups
- Repeating groups was removed
- If an object can have at most two groups, only make it have one and add the other one to the next object
- This is because if it only has two groups it is interpreted as a float and ultimately corrupts the image
- Added compression directly inside here for uploading - instead of relying on Geometry Dash to compress it
"""
from gdparse import GDLevel, LevelObject
from pathlib import Path
from .compression import compress_data, decompress_data
import gzip
import base64

BLOCK_ID = 211
GROUPS_PER_OBJECT = 10  # GD truncates groups beyond 10 when saving in the editor!


def encode(filepath: str | Path, skip_compression: bool = False) -> str:
    # Let's instead process 8 bytes at a time
    filepath = Path(filepath)
    file_data = filepath.read_bytes()

    # Prepend filename (1 byte length + filename bytes) before compression
    filename = filepath.name.encode('utf-8')
    if len(filename) > 255:
        filename = filename[:255]
    data = bytes([len(filename)]) + filename + file_data

    if not skip_compression:
        data = compress_data(data)
    all_groups = []
    # Process 8 bytes at a time
    for i in range(0, len(data), 8):
        chunk = data[i:i+8]

        if len(chunk) < 8:
            chunk = chunk.ljust(8, b'\x00')

        num = int.from_bytes(chunk, 'big')
        chunk_groups = []
        while num > 0:
            chunk_groups.append((num % 9999) + 1)  # 1-9999 instead of 0-9999
            num //= 9999
        # Pad to exactly 5 groups (8 bytes = max 5 base-9999 digits)
        while len(chunk_groups) < 5:
            chunk_groups.append(1)  # Use 1 as padding (represents 0)
        all_groups.extend(reversed(chunk_groups))

    # Store original length as first 2 groups (base 9999, supports up to ~99MB)
    # This keeps all group values within 1-9999
    length = len(data)
    len_high = (length // 9999) + 1  # High part (1-9999)
    len_low = (length % 9999) + 1    # Low part (1-9999)
    all_groups = [len_high, len_low] + all_groups

    level = GDLevel.create_empty()
    obj_index = 0
    current_obj_groups = []
    current_obj_set = set()  # Track groups in current object to avoid duplicates

    for group in all_groups:
        # If this group already exists in current object, or we hit the limit, start new object
        if group in current_obj_set or len(current_obj_groups) >= GROUPS_PER_OBJECT:
            # Would this create a 2-group object? (GD parses "X.Y" as float and corrupts it)
            if len(current_obj_groups) == 2:
                # Only save first group, push second to next object
                first_group = current_obj_groups[0]
                second_group = current_obj_groups[1]

                groups_str = str(first_group)
                obj = LevelObject.create_block(block_id=BLOCK_ID, x=obj_index * 30, y=0)
                obj.properties[57] = groups_str
                level.add_object(obj)
                obj_index += 1

                # Start new object with the pushed second group
                current_obj_groups = [second_group]
                current_obj_set = {second_group}
            else:
                # Save current object normally
                if current_obj_groups:
                    groups_str = '.'.join(str(g) for g in current_obj_groups)
                    obj = LevelObject.create_block(block_id=BLOCK_ID, x=obj_index * 30, y=0)
                    obj.properties[57] = groups_str
                    level.add_object(obj)
                    obj_index += 1
                # Start new object
                current_obj_groups = []
                current_obj_set = set()

        # Check again - the pushed group might conflict with the new group
        if group in current_obj_set:
            # Save the single pushed group and start fresh
            if current_obj_groups:
                groups_str = '.'.join(str(g) for g in current_obj_groups)
                obj = LevelObject.create_block(block_id=BLOCK_ID, x=obj_index * 30, y=0)
                obj.properties[57] = groups_str
                level.add_object(obj)
                obj_index += 1
            current_obj_groups = []
            current_obj_set = set()

        current_obj_groups.append(group)
        current_obj_set.add(group)

    # Don't forget the last object
    if current_obj_groups:
        groups_str = '.'.join(str(g) for g in current_obj_groups)
        obj = LevelObject.create_block(block_id=BLOCK_ID, x=obj_index * 30, y=0)
        obj.properties[57] = groups_str
        level.add_object(obj)

    # Serialize and compress to GD's expected format (gzip + base64)
    raw_level = level.serialize()
    compressed = gzip.compress(raw_level.encode('utf-8'))
    return base64.urlsafe_b64encode(compressed).decode('ascii').rstrip('=')

# 8 byte processing is way faster
# Faster rendering & smaller (~13%)
# Also, group 0 doesn't exist. I don't really use the GD Editor...


def decode(level_string: str, skip_decompression: bool = False) -> tuple[str, bytes]:
    """Decode a level string back to (filename, data)."""
    # Handle both compressed (H4sI...) and raw (kS38...) formats
    if level_string.startswith('H4sI'):
        # Gzip + base64 compressed format
        compressed = base64.urlsafe_b64decode(level_string + '==')
        level_string = gzip.decompress(compressed).decode('utf-8')

    level = GDLevel(level_string)

    # Collect all groups from all objects
    all_groups = []
    for obj in level.objects:
        groups_val = obj.properties.get(57)
        if groups_val is None:
            continue
        # Handle both string and numeric values
        groups_str = str(groups_val)
        all_groups.extend(int(g) for g in groups_str.split('.'))

    # Validate minimum data
    if len(all_groups) < 2:
        raise ValueError("Invalid level: not enough data to decode")

    # First 2 groups are the original data length (base 9999)
    len_high = all_groups[0]
    len_low = all_groups[1]
    original_len = (len_high - 1) * 9999 + (len_low - 1)
    all_groups = all_groups[2:]

    # Process 5 groups at a time (each 8 bytes = ~5 groups)
    result = bytearray()
    for i in range(0, len(all_groups), 5):
        chunk_groups = all_groups[i:i+5]

        # Convert base 9999 back to integer (groups are 1-9999, subtract 1)
        num = 0
        for g in chunk_groups:
            num = num * 9999 + (g - 1)

        result.extend(num.to_bytes(8, 'big'))

    # Trim to original length (remove padding)
    result = bytes(result[:original_len])

    if not skip_decompression:
        result = decompress_data(result)

    # Validate minimum data for filename extraction
    if len(result) < 1:
        raise ValueError("Invalid data: empty result after decompression")

    # Extract filename (1 byte length + filename bytes)
    filename_len = result[0]
    if len(result) < 1 + filename_len:
        raise ValueError("Invalid data: truncated filename")

    filename = result[1:1 + filename_len].decode('utf-8', errors='replace')
    file_data = result[1 + filename_len:]

    return filename, file_data

# ^ This is for the group method using Base9999 and 8 byte processing



"""
Method 1: X/Y Coordinates

Encodes each byte pair as X and Y coordinates of objects.
- First byte = X position
- Second byte = Y position
- If odd number of bytes, last object has Y = -1

While this did technically work - I didn't bother actually testing it as it was too unoptimized
"""
from gdparse import GDLevel, LevelObject
from pathlib import Path
from .compression import compress_data, decompress_data

BLOCK_ID = 211


def encode(filepath: str | Path, skip_compression: bool = False) -> str:
    filepath = Path(filepath)
    raw_data = filepath.read_bytes()

    compressed = compress_data(raw_data) if not skip_compression else raw_data

    print(f"Original: {len(raw_data)} bytes")
    print(f"Compressed: {len(compressed)} bytes ({len(compressed)/len(raw_data)*100:.1f}%)")

    # Create empty level
    level = GDLevel.create_empty()

    # Encode each byte as an object
    # Go through the bytes two by two - X being the first byte and y being the second
    for i in range(0, len(compressed) - len(compressed) % 2, 2):
        byte1 = compressed[i]
        byte2 = compressed[i+1]
        obj = LevelObject.create_block(
            block_id=BLOCK_ID,
            x=byte1,
            y=byte2,
        )
        level.add_object(obj)
    # Check if the last byte doesn't have a pair - in which case it has -1 as the y for the decoder
    if len(compressed) % 2 == 1:
        obj = LevelObject.create_block(
            block_id=BLOCK_ID,
            x=compressed[len(compressed) - 1],
            y=-1
        )
        level.add_object(obj)

    print(f"Objects created: {len(compressed)}")

    return level.serialize()

# While this did technically work - I didn't bother actually testing it as it was too unoptimized


def decode(level_string: str, skip_decompression: bool = False) -> bytes:

    level = GDLevel(level_string)

    compressed_bytes = []

    # Extract byte value from X and Y

    for obj in level.objects:
        x = obj.properties.get(2, 0)
        y = obj.properties.get(3, 0)

        compressed_bytes.append(x)

        if y != -1:
            compressed_bytes.append(y)

    compressed = bytes(compressed_bytes)

    # Decompress
    if not skip_decompression:
        return decompress_data(compressed)
    return compressed

#  ^ This is for the unoptimized version



"""
GD Save Manager - Fixed version of PyCCManager

Handles reading/writing Geometry Dash save files on Windows and macOS.

Fixes self reference error
"""
import os
import zlib
import base64
import struct
import platform
from Cryptodome.Cipher import AES


class GDData:
    def __init__(self, path):
        self.path = path
        self.ccll_path = f"{path}/CCLocalLevels.dat"
        self.ccgm_path = f"{path}/CCGameManager.dat"
        self.ccll = self.decode(open(self.ccll_path, "rb").read())
        self.ccgm = self.decode(open(self.ccgm_path, "rb").read())

    def injectLevel(self, levelData, levelName="Injected", levelDesc="Injected level"):
        levels = self.ccll.split(b">k_")
        header = levels[0]
        payload = (
            header + b">k_0</k><d><k>kCEK</k><i>4</i><k>k18</k><i>2</i><k>k2</k><s>"
            + levelName.encode() + b"</s><k>k4</k><s>"
            + levelData.encode() + b"</s><k>k5</k><s>"
            + levelDesc.encode() + b"</s><k>k13</k><t /><k>k21</k><i>2</i><k>k16</k><i>1</i>"
            + b"<k>k80</k><i>338</i><k>k81</k><i>23</i><k>k83</k><i>109</i><k>k50</k><i>35</i>"
            + b"<k>k48</k><i>23</i><k>kI1</k><r>-1118.36</r><k>kI2</k><r>-366.449</r>"
            + b"<k>kI3</k><r>0.7</r><k>kI4</k><i>2</i><k>kI5</k><i>11</i><k>kI7</k><i>1</i>"
            + b"<k>kI6</k><d><k>0</k><s>0</s><k>1</k><s>0</s><k>2</k><s>0</s><k>3</k><s>0</s>"
            + b"<k>4</k><s>0</s><k>5</k><s>0</s><k>6</k><s>0</s><k>7</k><s>0</s><k>8</k><s>0</s>"
            + b"<k>9</k><s>0</s><k>10</k><s>0</s><k>11</k><s>2</s><k>12</k><s>0</s></d></d><k"
        )
        for i in range(1, len(levels)):
            payload += b">k_" + str(i).encode() + b"<" + b"<".join(levels[i].split(b"<")[1:])
        self.ccll = payload

    def save(self, ccll=True, ccgm=True):
        if ccll:
            open(self.ccll_path, "wb").write(self.encode(self.ccll))
        if ccgm:
            open(self.ccgm_path, "wb").write(self.encode(self.ccgm))

    def encode(self, data):
        raise NotImplementedError

    def decode(self, data):
        raise NotImplementedError


class GDWinData(GDData):
    """Windows GD save format (XOR + zlib)"""

    def __init__(self, path=None):
        if path is None:
            path = os.path.join(os.getenv("LOCALAPPDATA", ""), "GeometryDash")
        super().__init__(path)

    def encode(self, data):
        compressed = zlib.compress(data)
        crc32 = struct.pack('I', zlib.crc32(data))
        size = struct.pack('I', len(data))
        encrypted = base64.b64encode(
            b'\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x0b' + compressed[2:-4] + crc32 + size,
            b'-_'
        )
        return bytes([a ^ 11 for a in encrypted])

    def decode(self, data):
        decrypted = bytes([a ^ 11 for a in data])
        return zlib.decompress(base64.b64decode(decrypted, b'-_')[10:], -zlib.MAX_WBITS)


class GDMacData(GDData):
    """macOS GD save format (AES encryption)"""

    MAC_KEY = b"\x69\x70\x75\x39\x54\x55\x76\x35\x34\x79\x76\x5d\x69\x73\x46\x4d\x68\x35\x40\x3b\x74\x2e\x35\x77\x33\x34\x45\x32\x52\x79\x40\x7b"

    def __init__(self, path=None):
        if path is None:
            path = os.path.join(os.path.expanduser("~"), "Library/Application Support/GeometryDash")
        self.cipher = AES.new(GDMacData.MAC_KEY, AES.MODE_ECB)
        super().__init__(path)

    def encode(self, data):
        extra = len(data) % 16
        if extra > 0:
            data += (b'\x0b' * (16 - extra))
        return self.cipher.encrypt(data)

    def decode(self, data):
        return self.cipher.decrypt(data)


def new_manager(path=None, format="auto"):
    """
    Create a GD save manager.

    Args:
        path: Custom GD save folder path. If None, uses platform default.
        format: "auto", "windows", or "mac"
    """
    if format == "auto":
        format = "mac" if platform.system() == "Darwin" else "windows"

    if format == "mac":
        return GDMacData(path)
    else:
        return GDWinData(path)



"""
Method 5: Property 31 (Text Property)

Stores base64 encoded data in property 31 (text property) of a single object.

Now I didn't know why I thought this would work but - it didn't...
"""
from gdparse import GDLevel, LevelObject
from pathlib import Path
import base64
from .compression import compress_data, decompress_data

BLOCK_ID = 211


def encode(filepath: str | Path, skip_compression: bool = False) -> str:
    filepath = Path(filepath)
    data = filepath.read_bytes()
    if not skip_compression:
        data = compress_data(data)

    # Encode file in Base64
    b64_string = base64.b64encode(data).decode('ascii')
    level = GDLevel.create_empty()

    # Create a single object
    obj = LevelObject.create_block(block_id=BLOCK_ID, x=0, y=0)

    # Add the whole data inside a single objects string
    obj.properties[31] = b64_string
    level.add_object(obj)

    return level.serialize()

# Now I didn't know why I thought this would work but - it didn't...


def decode(level_string: str, skip_decompression: bool = False) -> bytes:
    level = GDLevel(level_string)

    b64_string = level.objects[0].properties.get(31)
    data = base64.b64decode(b64_string)

    if not skip_decompression:
        data = decompress_data(data)

    return data

# ^ This is for the (stupid) text property version
